"""LICENSE
Copyright 2017 Hermann Krumrey <hermann@krumreyh.com>

This file is part of bundesliga-tippspiel.

bundesliga-tippspiel is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

bundesliga-tippspiel is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with bundesliga-tippspiel.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

# noinspection PyProtectedMember
from unittest import TestCase
from puffotter.units import human_readable_bytes


class TestCrypto(TestCase):
    """
    Tests cryptographical functions
    """

    def test_converting_bytes_to_human_readable(self):
        """
        Tests that passwords can be hashed successfully
        :return: None
        """
        self.assertEqual("1MB", human_readable_bytes(1000000))
        self.assertEqual("1.024KB", human_readable_bytes(1024))
        self.assertEqual("0.123KB", human_readable_bytes(123))
        self.assertEqual("1.234GB", human_readable_bytes(1234123123))



"""LICENSE
Copyright 2017 Hermann Krumrey <hermann@krumreyh.com>

This file is part of bundesliga-tippspiel.

bundesliga-tippspiel is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

bundesliga-tippspiel is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with bundesliga-tippspiel.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

# noinspection PyProtectedMember
import os
import shutil
from unittest import TestCase
from puffotter.os import listdir


class TestCrypto(TestCase):
    """
    Tests os functions
    """

    def test_listdir(self):
        """
        Tests the listdir function
        :return: None
        """
        os.makedirs("test")
        os.makedirs("test/1")
        os.makedirs("test/.2")
        open("test/3", "w").close()
        open("test/.4", "w").close()

        _all = listdir("test", no_dot=False)
        self.assertEqual(len(_all), 4)

        no_dot = listdir("test")
        self.assertEqual(len(no_dot), 2)

        no_files = listdir("test", no_files=True)
        self.assertEqual(len(no_files), 1)
        self.assertEqual(no_files[0], ("1", "test/1"))

        no_dirs = listdir("test", no_dirs=True)
        self.assertEqual(len(no_dirs), 1)
        self.assertEqual(no_dirs[0], ("3", "test/3"))

        nothing = listdir("test", no_files=True, no_dirs=True)
        self.assertEqual(len(nothing), 0)

        shutil.rmtree("test")



"""LICENSE
Copyright 2019 Hermann Krumrey <hermann@krumreyh.com>

This file is part of puffotter.

puffotter is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

puffotter is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with puffotter.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

import sys
import base64
import binascii
import logging
import sentry_sdk
import traceback
from logging.handlers import TimedRotatingFileHandler
from sqlalchemy.exc import OperationalError
from sentry_sdk.integrations.logging import LoggingIntegration
from sentry_sdk.integrations.flask import FlaskIntegration
from typing import List, Optional, Type, Callable, Tuple
from flask import redirect, url_for, flash, render_template
from flask.logging import default_handler
from flask.blueprints import Blueprint
from werkzeug.exceptions import HTTPException
from puffotter.flask.Config import Config
from puffotter.flask.base import app, login_manager, db
from puffotter.flask.enums import AlertSeverity
from puffotter.flask.db.User import User
from puffotter.flask.db.ApiKey import ApiKey
from puffotter.flask.db.TelegramChatId import TelegramChatId
from puffotter.flask.routes import blueprint_generators \
    as default_blueprint_generators


CREATED_BLUEPRINTS = []
"""
Keeps track of created blueprint names.
This is necessary for unit testing with nose, because duplicate blueprint names
will cause errors.
"""


def init_flask(
        module_name: str,
        sentry_dsn: str,
        root_path: str,
        config: Type[Config],
        models: List[Type[db.Model]],
        blueprint_generators: List[Tuple[Callable[[str], Blueprint], str]]
):
    """
    Initializes the flask application
    :param module_name: The name of the module
    :param sentry_dsn: The sentry DSN used for error logging
    :param root_path: The root path of the flask application
    :param config: The Config class to use for configuration
    :param models: The database models to create
    :param blueprint_generators: Tuples that contain a function that generates
                                 a blueprint and the name of the blueprint
    :return: None
    """
    app.root_path = root_path
    config.load_config(root_path, module_name, sentry_dsn)
    __init_logging(config)

    default_models = [
        User,
        ApiKey,
        TelegramChatId
    ]

    __init_app(config, default_blueprint_generators + blueprint_generators)
    __init_db(config, default_models + models)
    __init_login_manager()


def __init_logging(config: Type[Config]):
    """
    Sets up logging to a logfile
    :param config: The configuration to use
    :return: None
    """
    sentry_logging = LoggingIntegration(
        level=logging.INFO,
        event_level=None
    )
    sentry_sdk.init(
        dsn=config.SENTRY_DSN,
        integrations=[FlaskIntegration(), sentry_logging]
    )

    app.logger.removeHandler(default_handler)

    log_format = \
        "[%(asctime)s, %(levelname)s] %(module)s[%(lineno)d]: %(message)s"
    formatter = logging.Formatter(log_format)

    info_handler = TimedRotatingFileHandler(
        config.LOGGING_PATH,
        when="midnight",
        interval=1,
        backupCount=7
    )
    info_handler.setLevel(logging.INFO)
    info_handler.setFormatter(formatter)

    debug_handler = TimedRotatingFileHandler(
        config.DEBUG_LOGGING_PATH,
        when="midnight",
        interval=1,
        backupCount=7
    )
    debug_handler.setLevel(logging.DEBUG)
    debug_handler.setFormatter(formatter)

    stream_handler = logging.StreamHandler(stream=sys.stdout)
    stream_handler.setLevel(config.VERBOSITY)
    stream_handler.setFormatter(formatter)

    app.logger.addHandler(info_handler)
    app.logger.addHandler(debug_handler)
    app.logger.addHandler(stream_handler)

    app.logger.setLevel(logging.DEBUG)


def __init_app(
        config: Type[Config],
        blueprint_generators: List[Tuple[Callable[[str], Blueprint], str]]
):
    """
    Initializes the flask app
    :param config: The configuration to use
    :param blueprint_generators: Tuples that contain a function that generates
                                 a blueprint and the name of the blueprint
    :return: None
    """
    app.testing = config.TESTING
    app.config["TRAP_HTTP_EXCEPTIONS"] = True
    app.config["SERVER_NAME"] = Config.base_url().split("://", 1)[1]
    app.secret_key = config.FLASK_SECRET
    for blueprint_generator, blueprint_name in blueprint_generators:
        if blueprint_name in CREATED_BLUEPRINTS:
            app.logger.debug(f"Blueprint {blueprint_name} already created")
            continue
        else:
            app.logger.info(f"Creating blueprint {blueprint_name}")
            CREATED_BLUEPRINTS.append(blueprint_name)
            blueprint = blueprint_generator(blueprint_name)
            app.register_blueprint(blueprint)

    @app.context_processor
    def inject_template_variables():
        """
        Injects the project's version string so that it will be available
        in templates
        :return: The dictionary to inject
        """
        return {
            "version": config.VERSION,
            "env": app.env,
            "config": config
        }

    @app.errorhandler(Exception)
    def exception_handling(e: Exception):
        """
        Handles any uncaught exceptions and shows an applicable error page
        :param e: The caught exception
        :return: The response to the exception
        """
        if isinstance(e, HTTPException):
            error = e
            if e.code == 401:
                flash(
                    config.STRINGS["401_message"],
                    AlertSeverity.DANGER.value
                )
                return redirect(url_for("user_management.login"))
            app.logger.warning("Caught HTTP exception: {}".format(e))
        else:
            error = HTTPException(config.STRINGS["500_message"])
            error.code = 500
            trace = "".join(traceback.format_exception(*sys.exc_info()))
            app.logger.error("Caught exception: {}\n{}".format(e, trace))
            sentry_sdk.capture_exception(e)
        return render_template(
            config.REQUIRED_TEMPLATES["error_page"],
            error=error
        )

    @app.errorhandler(HTTPException)
    def unauthorized_handling(e: HTTPException):
        """
        Forwards HTTP exceptions to the error handler
        :param e: The HTTPException
        :return: The response to the exception
        """
        return exception_handling(e)


def __init_db(config: Type[Config], models: List[db.Model]):
    """
    Initializes the database
    :param config: The configuration to use
    :param models: The models to create in the database
    :return: None
    """
    app.config["SQLALCHEMY_DATABASE_URI"] = config.DB_URI
    app.config["SQLALCHEMY_TRACK_MODIFICATIONS"] = False

    # Makes sure that we don't get errors because
    # of an idle database connection
    app.config["SQLALCHEMY_ENGINE_OPTIONS"] = {"pool_pre_ping": True}

    db.init_app(app)

    for model in models:
        app.logger.debug(f"Loading model {model.__name__}")

    with app.app_context():
        try:
            db.create_all()
        except OperationalError:
            print("Failed to connect to the database")
            sys.exit(1)


def __init_login_manager():
    """
    Initializes the login manager
    :return: None
    """
    login_manager.session_protection = "strong"

    # Set up login manager
    @login_manager.user_loader
    def load_user(user_id: str) -> Optional[User]:
        """
        Loads a user from an ID
        :param user_id: The ID
        :return: The User
        """
        return User.query.get(int(user_id))

    @login_manager.request_loader
    def load_user_from_request(request) -> Optional[User]:
        """
        Loads a user pased on a provided API key
        :param request: The request containing the API key in the headers
        :return: The user or None if no valid API key was provided
        """
        if "Authorization" not in request.headers:
            return None

        api_key = request.headers["Authorization"].replace("Basic ", "", 1)

        try:
            api_key = base64.b64decode(
                api_key.encode("utf-8")
            ).decode("utf-8")
        except (TypeError, binascii.Error):
            return None

        db_api_key = ApiKey.query.get(api_key.split(":", 1)[0])

        # Check for validity of API key
        if db_api_key is None or not db_api_key.verify_key(api_key):
            return None

        elif db_api_key.has_expired():
            db.session.delete(db_api_key)
            db.session.commit()
            return None

        return User.query.get(db_api_key.user_id)



"""LICENSE
Copyright 2019 Hermann Krumrey <hermann@krumreyh.com>

This file is part of puffotter.

puffotter is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

puffotter is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with puffotter.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

import time
import requests
import logging


def aggressive_request(url: str) -> str:
    """
    Handles GET requests while analyzing status codes
    :param url: The URL to get
    :return: The response text
    """
    logger = logging.getLogger("puffotter")

    time.sleep(1)
    headers = {"User-Agent": "Mozilla/5.0"}
    resp = requests.get(url, headers=headers)

    while resp.status_code >= 300:
        logger.warning("HTTP Error Code: {}".format(resp.status_code))
        logger.debug(resp.headers)
        logger.debug(resp.text)
        time.sleep(60)
        resp = requests.get(url, headers=headers)

    return resp.text



"""LICENSE
Copyright 2019 Hermann Krumrey <hermann@krumreyh.com>

This file is part of puffotter.

puffotter is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

puffotter is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with puffotter.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

import os
import shutil
from base64 import b64encode
from unittest import TestCase
from typing import Tuple, Dict, List, Type, Callable
from flask.blueprints import Blueprint
from flask_login import login_user
from puffotter.crypto import generate_random, generate_hash
from puffotter.flask.Config import Config
from puffotter.flask.db.User import User
from puffotter.flask.db.ApiKey import ApiKey
from puffotter.flask.initialize import init_flask, app, db
from puffotter.env import load_env_file


class _TestFramework(TestCase):
    """
    Class that models a testing framework for the flask application
    """

    GENERATED_USER_CREDENTIALS: List[Tuple[str, str, str, str]] = []
    """
    List of dynamically generated generated user credentials
    (password, password hash, confirmation key, confirmation hash)
    Used to increase speed of the tests, since hashing with bcrypt takes a
    relatively long time
    """

    GENERATED_API_KEY_CREDENTIALS: List[Tuple[str, str]] = []
    """
    List of dynamically generated API keys and their hashes.
    Used to speed up the testing
    """

    maxDiff = None

    module_name: str = "puffotter"
    models: List[db.Model] = []
    blueprint_generators: List[Tuple[Callable[[str], Blueprint], str]] = []
    root_path: str = "."
    config: Type[Config] = Config

    def setUp(self):
        """
        Sets up the flask application and a temporary database to test with
        :return: None
        """
        self.user_cred_count = 0
        self.api_cred_count = 0

        self.temp_templates_dir = "templates"
        self.templates_sample_dir = os.path.join(
            os.path.dirname(os.path.abspath(__file__)),
            "templates"
        )
        os.environ["FLASK_TESTING"] = "1"
        os.environ["DB_MODE"] = "sqlite"
        os.environ["FLASK_SECRET"] = generate_random(20)

        self.cleanup(True)
        if self.module_name == "puffotter":
            shutil.copytree(self.templates_sample_dir, self.temp_templates_dir)
            os.environ["RECAPTCHA_SITE_KEY"] = ""
            os.environ["RECAPTCHA_SECRET_KEY"] = ""
            os.environ["SMTP_HOST"] = ""
            os.environ["SMTP_PORT"] = "0"
            os.environ["SMTP_ADDRESS"] = ""
            os.environ["SMTP_PASSWORD"] = ""
            os.environ["TELEGRAM_API_KEY"] = ""
        else:
            load_env_file()

        self.config.load_config(self.root_path, self.module_name, "")
        self.db_path = Config.DB_URI.split("sqlite:///")[1]
        self.cleanup(False)

        self.app = app
        self.db = db

        init_flask(
            self.module_name,
            "",
            self.root_path,
            self.config,
            self.models,
            self.blueprint_generators
        )
        self.app.app_context().push()

        self.client = self.app.test_client()
        self.context = self.app.test_request_context()

    def tearDown(self):
        """
        Removes any generated files from the filesystem and handles other
        post-test tasks
        :return: None
        """
        self.cleanup(True)

    def cleanup(self, remove_templates: bool = True):
        """
        Cleans up the filesystem after/before tests
        :return: None
        """
        try:
            if self.db_path != "" and os.path.isfile(self.db_path):
                os.remove(self.db_path)
        except AttributeError:
            pass
        if os.path.isdir(self.temp_templates_dir) and remove_templates:
            shutil.rmtree(self.temp_templates_dir)

    def generate_sample_user(self, confirmed: bool = True) \
            -> Tuple[User, str, str]:
        """
        Generates a random user for use in tests
        :param confirmed: Whether or not the user should be confirmed
        :return: The User object, the password and the confirmation key
        """
        if self.user_cred_count >= \
                len(_TestFramework.GENERATED_USER_CREDENTIALS):
            password = generate_random(20)
            password_hash = generate_hash(password)
            confirm_key = generate_random(20)
            confirmation_hash = generate_hash(confirm_key)
            _TestFramework.GENERATED_USER_CREDENTIALS.append(
                (password, password_hash, confirm_key, confirmation_hash)
            )
        else:
            password, password_hash, confirm_key, confirmation_hash = \
                _TestFramework.GENERATED_USER_CREDENTIALS[self.user_cred_count]
        self.user_cred_count += 1

        user = User(
            username=generate_random(12),
            password_hash=password_hash,
            email=generate_random(8) + "@example.com",
            confirmed=confirmed,
            confirmation_hash=confirmation_hash
        )
        self.db.session.add(user)
        self.db.session.commit()
        return user, password, confirm_key

    def login_user(self, user: User, password: str, use_client: bool = True):
        """
        Logs in a user
        :param user: The user to log in
        :param password: The password to use
        :param use_client: Whether or not to use the testing clienty
        :return: None
        """
        if use_client:
            self.client.post("/login", follow_redirects=True, data={
                "username": user.username,
                "password": password
            })
        else:
            applicable_hashes = [
                x[1]
                for x in _TestFramework.GENERATED_USER_CREDENTIALS
                if x[0] == password
            ]
            if user.password_hash not in applicable_hashes:
                return
            else:
                with self.context:
                    login_user(user)

    def generate_api_key(self, user: User) \
            -> Tuple[ApiKey, str, Dict[str, str]]:
        """
        Generates an API key and base64 encoded headers for requests
        :param user: The user for which to create the key
        :return: The API key object, the actual API key, the headers
        """
        if self.api_cred_count >= \
                len(_TestFramework.GENERATED_API_KEY_CREDENTIALS):
            key = generate_random(20)
            hashed = generate_hash(key)
            _TestFramework.GENERATED_API_KEY_CREDENTIALS.append((key, hashed))
        else:
            key, hashed = _TestFramework.GENERATED_API_KEY_CREDENTIALS[
                self.api_cred_count
            ]
        self.api_cred_count += 1

        api_key_obj = ApiKey(user=user, key_hash=hashed)
        self.db.session.add(api_key_obj)
        self.db.session.commit()
        api_key = "{}:{}".format(api_key_obj.id, key)

        return api_key_obj, api_key, self.generate_api_key_headers(api_key)

    # noinspection PyMethodMayBeStatic
    def generate_api_key_headers(self, api_key: str) -> Dict[str, str]:
        """
        Creates HTTP Authorization headers for an API key
        :param api_key: The API key to put in the headers
        :return: The headers
        """
        encoded = b64encode(api_key.encode("utf-8")).decode("utf-8")
        return {
            "Authorization": "Basic {}".format(encoded)
        }



"""LICENSE
Copyright 2019 Hermann Krumrey <hermann@krumreyh.com>

This file is part of puffotter.

puffotter is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

puffotter is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with puffotter.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

import time
from puffotter.flask.Config import Config
from puffotter.flask.db.ApiKey import ApiKey
from puffotter.flask.test.TestFramework import _TestFramework


class TestApiKey(_TestFramework):
    """
    Class that tests the ApiKey database model
    """

    def test_json_representation(self):
        """
        Tests the JSON representation of the model
        :return: None
        """
        user = self.generate_sample_user()[0]
        api_key = self.generate_api_key(user)[0]
        self.assertEqual(
            api_key.__json__(False),
            {
                "id": api_key.id,
                "user_id": user.id,
                "creation_time": api_key.creation_time
            }
        )
        user_json = user.__json__(True)
        user_json.pop("api_keys")
        self.assertEqual(
            api_key.__json__(True),
            {
                "id": api_key.id,
                "user_id": user.id,
                "user": user_json,
                "creation_time": api_key.creation_time
            }
        )

    def test_string_representation(self):
        """
        Tests the string representation of the model
        :return: None
        """
        user = self.generate_sample_user()[0]
        api_key = self.generate_api_key(user)[0]
        data = api_key.__json__()
        data.pop("id")
        self.assertEqual(
            str(api_key),
            "ApiKey:{} <{}>".format(api_key.id, data)
        )

    def test_repr(self):
        """
        Tests the __repr__ method of the model class
        :return: None
        """
        user = self.generate_sample_user()[0]
        api_key = self.generate_api_key(user)[0]
        generated = {"value": api_key}
        code = repr(api_key)

        exec("generated['value'] = " + code)
        self.assertEqual(generated["value"], api_key)
        self.assertFalse(generated["value"] is api_key)

    def test_hashing(self):
        """
        Tests using the model objects as keys in a dictionary
        :return: None
        """
        user = self.generate_sample_user()[0]
        api_key = self.generate_api_key(user)[0]
        api_key2 = self.generate_api_key(user)[0]
        mapping = {
            api_key: 100,
            api_key2: 200
        }
        self.assertEqual(mapping[api_key], 100)
        self.assertEqual(mapping[api_key2], 200)

    def test_equality(self):
        """
        Tests checking equality for model objects
        :return: None
        """
        user = self.generate_sample_user()[0]
        api_key = self.generate_api_key(user)[0]
        api_key2 = self.generate_api_key(user)[0]
        self.assertEqual(api_key, api_key)
        self.assertNotEqual(api_key, api_key2)
        self.assertNotEqual(api_key, 100)

    def test_verifying_key(self):
        """
        Tests verifying an api key
        :return: None
        """
        user = self.generate_sample_user()[0]
        api_key, key, _ = self.generate_api_key(user)
        self.assertTrue(api_key.verify_key(key))
        self.assertFalse(api_key.verify_key("A:1"))
        self.assertFalse(api_key.verify_key("100:" + key.split(":", 1)[1]))
        self.assertFalse(api_key.verify_key("{}:ABC".format(api_key.id)))

    def test_expiration(self):
        """
        Tests if the expiration of API keys works correctly
        :return: None
        """
        self.assertTrue(ApiKey(creation_time=0).has_expired())
        self.assertTrue(ApiKey(
            creation_time=time.time() - Config.MAX_API_KEY_AGE
        ).has_expired())
        self.assertFalse(ApiKey(
            creation_time=time.time() - Config.MAX_API_KEY_AGE + 1
        ).has_expired())
        self.assertFalse(ApiKey(creation_time=time.time()).has_expired())
        backup = Config.MAX_API_KEY_AGE
        Config.MAX_API_KEY_AGE = 0
        self.assertTrue(ApiKey(creation_time=time.time()).has_expired())
        Config.MAX_API_KEY_AGE = backup



"""LICENSE
Copyright 2019 Hermann Krumrey <hermann@krumreyh.com>

This file is part of puffotter.

puffotter is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

puffotter is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with puffotter.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

from enum import Enum
from typing import Dict, Any
from puffotter.flask.base import db
from puffotter.flask.db.ModelMixin import ModelMixin
from puffotter.flask.test.TestFramework import _TestFramework


class TestModelMixin(_TestFramework):
    """
    Class that tests the ModelMixin class
    """

    def test_enum_attributes(self):
        """
        Tests if the repr method handles enums correctly
        :return: None
        """

        class A(Enum):
            B = 1
            C = 2

        class Tester(ModelMixin, db.Model):
            enum = db.Column(db.Enum(A))

            def __init__(self, *args, **kwargs):
                super().__init__(*args, **kwargs)

            def __json__(self, include_children: bool = False,  _=None) \
                    -> Dict[str, Any]:
                return {"id": self.id, "enum": self.enum.value}

        tester = Tester(id=1, enum=A.B)
        self.assertEqual(repr(tester), "Tester(id=1, enum=A.B)")



"""LICENSE
Copyright 2019 Hermann Krumrey <hermann@krumreyh.com>

This file is part of puffotter.

puffotter is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

puffotter is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with puffotter.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

from puffotter.flask.db.TelegramChatId import TelegramChatId
from puffotter.flask.test.TestFramework import _TestFramework


class TestUser(_TestFramework):
    """
    Class that tests the User database model
    """

    def test_json_representation(self):
        """
        Tests the JSON representation of the model
        :return: None
        """
        user = self.generate_sample_user()[0]
        should = {
            "id": user.id,
            "email": user.email,
            "confirmed": True,
            "username": user.username
        }
        self.assertEqual(user.__json__(False), should)
        should.update({
            "telegram_chat_id": None,
            "api_keys": []
        })
        self.assertEqual(user.__json__(True), should)

    def test_string_representation(self):
        """
        Tests the string representation of the model
        :return: None
        """
        user = self.generate_sample_user()[0]
        data = user.__json__()
        data.pop("id")
        self.assertEqual(
            str(user),
            "User:{} <{}>".format(user.id, data)
        )

    def test_repr(self):
        """
        Tests the __repr__ method of the model class
        :return: None
        """
        user = self.generate_sample_user()[0]
        generated = {"value": user}
        code = repr(user)

        # noinspection PyUnresolvedReferences
        from puffotter.flask.db.User import User

        exec("generated['value'] = " + code)
        self.assertEqual(generated["value"], user)
        self.assertFalse(generated["value"] is user)

    def test_hashing(self):
        """
        Tests using the model objects as keys in a dictionary
        :return: None
        """
        user = self.generate_sample_user()[0]
        user2 = self.generate_sample_user()[0]
        mapping = {
            user: 100,
            user2: 200
        }
        self.assertEqual(mapping[user], 100)
        self.assertEqual(mapping[user2], 200)

    def test_equality(self):
        """
        Tests checking equality for model objects
        :return: None
        """
        user = self.generate_sample_user()[0]
        user2 = self.generate_sample_user()[0]
        self.assertEqual(user, user)
        self.assertNotEqual(user, user2)
        self.assertNotEqual(user, 100)

    def test_flask_properties(self):
        """
        Tests if the flask_login properties work as expected
        :return: None
        """
        user = self.generate_sample_user()[0]
        self.assertEqual(user.is_authenticated, True)
        self.assertEqual(user.is_anonymous, False)
        self.assertEqual(user.is_active, user.confirmed)
        self.assertEqual(user.get_id(), str(user.id))

    def test_verifying_password(self):
        """
        Tests verifying the password of a user
        :return: None
        """
        user, password, confirm_key = self.generate_sample_user()
        self.assertTrue(user.verify_password(password))
        self.assertFalse(user.verify_password("AAAAAA"))
        self.assertTrue(user.verify_confirmation(confirm_key))
        self.assertFalse(user.verify_confirmation("AAAAAA"))

    def test_json_with_telegram_chat_id(self):
        """
        Tests whether or not including children can cause recursion errors
        :return: None
        """
        user, _, _ = self.generate_sample_user(True)
        chat_id = TelegramChatId(user=user, chat_id="AAA")
        api_key, _, _ = self.generate_api_key(user)
        self.db.session.add(chat_id)
        self.db.session.commit()
        should = user.__json__(False)
        should.update({
            "telegram_chat_id": chat_id.__json__(True),
            "api_keys": [api_key.__json__(True)]
        })
        should["telegram_chat_id"].pop("user")
        for x in should["api_keys"]:
            x.pop("user")
        self.assertEqual(user.__json__(True), should)



"""LICENSE
Copyright 2019 Hermann Krumrey <hermann@krumreyh.com>

This file is part of puffotter.

puffotter is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

puffotter is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with puffotter.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

from bokkichat.entities.Address import Address
from bokkichat.entities.message.TextMessage import TextMessage
from puffotter.flask.base import db
from puffotter.flask.Config import Config
from puffotter.flask.db.ModelMixin import ModelMixin
from puffotter.flask.db.User import User


class TelegramChatId(ModelMixin, db.Model):
    """
    Model that describes the 'telegram_chat_ids' SQL table
    Maps telegram chat ids to users
    """

    def __init__(self, *args, **kwargs):
        """
        Initializes the Model
        :param args: The constructor arguments
        :param kwargs: The constructor keyword arguments
        """
        super().__init__(*args, **kwargs)

    __tablename__ = "telegram_chat_ids"
    """
    The name of the table
    """

    user_id: int = db.Column(
        db.Integer, db.ForeignKey("users.id"),
        nullable=False
    )
    """
    The ID of the user associated with this telegram chat ID
    """

    user: User = db.relationship("User", back_populates="telegram_chat_id")
    """
    The user associated with this telegram chat ID
    """

    chat_id: str = db.Column(db.String(255), nullable=False)
    """
    The telegram chat ID
    """

    def send_message(self, message_text: str):
        """
        Sends a message to the telegram chat
        :param message_text: The message text to send
        :return: None
        """
        address = Address(self.chat_id)
        message = TextMessage(
            Config.TELEGRAM_BOT_CONNECTION.address,
            address,
            message_text
        )
        Config.TELEGRAM_BOT_CONNECTION.send(message)



"""LICENSE
Copyright 2019 Hermann Krumrey <hermann@krumreyh.com>

This file is part of puffotter.

puffotter is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

puffotter is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with puffotter.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

import smtplib
from email.mime.text import MIMEText
from email.mime.multipart import MIMEMultipart


def send_email(
        address: str,
        title: str,
        message: str,
        smtp_server: str,
        smtp_address: str,
        smtp_password: str,
        smtp_port: int = 587,
):
    """
    Sends an HTML email message using SMTP
    :param address: The address to send to
    :param title: The email's title
    :param message: The message to send
    :param smtp_server: The SMTP server to use
    :param smtp_address: The SMTP address to use
    :param smtp_password: The SMTP password to use
    :param smtp_port: The SMTP port to use
    :return: None
    """
    connection = smtplib.SMTP(smtp_server, smtp_port)
    connection.ehlo()
    connection.starttls()
    connection.ehlo()
    connection.login(smtp_address, smtp_password)

    msg = MIMEMultipart("alternative")
    msg["subject"] = title
    msg["From"] = smtp_address
    msg["To"] = address
    msg.attach(MIMEText(message, "html"))

    connection.sendmail(smtp_address, address, msg.as_string())
    connection.quit()



import os
from typing import Union
from werkzeug import Response
from flask import Blueprint, redirect, url_for, request, render_template, flash
from flask_login import login_required, current_user, logout_user, login_user
from puffotter.crypto import generate_hash, generate_random
from puffotter.recaptcha import verify_recaptcha
from puffotter.smtp import send_email
from puffotter.flask.base import app, db
from puffotter.flask.Config import Config
from puffotter.flask.db.User import User
from puffotter.flask.db.TelegramChatId import TelegramChatId


def define_blueprint(blueprint_name: str) -> Blueprint:
    """
    Defines the blueprint for this route
    :param blueprint_name: The name of the blueprint
    :return: The blueprint
    """
    blueprint = Blueprint(blueprint_name, __name__)

    @blueprint.route("/login", methods=["GET", "POST"])
    def login() -> Union[Response, str]:
        """
        Page that allows the user to log in
        :return: The response
        """
        if request.method == "POST":
            username = request.form["username"]
            password = request.form["password"]
            remember_me = request.form.get("remember_me") in ["on", True]

            user: User = User.query.filter_by(username=username).first()

            if user is None:
                flash(Config.STRINGS["user_does_not_exist"], "danger")
            elif current_user.is_authenticated:
                flash(Config.STRINGS["user_already_logged_in"], "info")
            elif not user.confirmed:
                flash(Config.STRINGS["user_is_not_confirmed"], "danger")
            elif not user.verify_password(password):
                flash(Config.STRINGS["invalid_password"], "danger")
            else:
                login_user(user, remember=remember_me)
                flash(Config.STRINGS["logged_in"], "success")
                app.logger.info(f"User {current_user.username} logged in.")
                return redirect(url_for("static.index"))
            return redirect(url_for("user_management.login"))
        else:
            return render_template(
                "user_management/login.html",
                **Config.TEMPLATE_EXTRAS["login"]()
            )

    @blueprint.route("/logout", methods=["GET"])
    @login_required
    def logout() -> Union[Response, str]:
        """
        Logs out the user
        :return: The response
        """
        app.logger.info("User {} logged out.".format(current_user.username))
        logout_user()
        flash(Config.STRINGS["logged_out"], "info")
        return redirect(url_for("static.index"))

    @blueprint.route("/register", methods=["GET", "POST"])
    def register() -> Union[Response, str]:
        """
        Page that allows a new user to register
        :return: The response
        """
        if request.method == "POST":
            username = request.form["username"]
            email = request.form["email"]
            password = request.form["password"]
            password_repeat = request.form["password-repeat"]
            recaptcha_result = verify_recaptcha(
                request.remote_addr,
                request.form.get("g-recaptcha-response", ""),
                Config.RECAPTCHA_SECRET_KEY
            )

            all_users = User.query.all()
            usernames = [user.username for user in all_users]
            emails = [user.email for user in all_users]

            _min, _max = Config.MIN_USERNAME_LENGTH, Config.MAX_USERNAME_LENGTH

            if len(username) < _min or len(username) > _max:
                flash(Config.STRINGS["username_length"]
                      .format(_min, _max), "danger")
            elif password != password_repeat:
                flash(Config.STRINGS["passwords_do_not_match"], "danger")
            elif username in usernames:
                flash(Config.STRINGS["username_already_exists"], "danger")
            elif email in emails:
                flash(Config.STRINGS["email_already_in_use"], "danger")
            elif not recaptcha_result:
                flash(Config.STRINGS["recaptcha_incorrect"], "danger")
            else:
                confirmation_key = generate_random(32)
                confirmation_hash = generate_hash(confirmation_key)
                user = User(
                    username=username,
                    email=email,
                    password_hash=generate_hash(password),
                    confirmation_hash=confirmation_hash
                )
                db.session.add(user)
                db.session.commit()
                email_msg = render_template(
                    "email/registration.html",
                    domain_name=Config.DOMAIN_NAME,
                    host_url=Config.base_url(),
                    target_url=os.path.join(Config.base_url(), "confirm"),
                    username=username,
                    user_id=user.id,
                    confirm_key=confirmation_key,
                    **Config.TEMPLATE_EXTRAS["registration_email"]()
                )
                send_email(
                    email,
                    Config.STRINGS["registration_email_title"],
                    email_msg,
                    Config.SMTP_HOST,
                    Config.SMTP_ADDRESS,
                    Config.SMTP_PASSWORD,
                    Config.SMTP_PORT
                )
                app.logger.info("User {} registered.".format(user.username))
                flash(Config.STRINGS["registration_successful"], "info")
                return redirect(url_for("static.index"))
            return redirect(url_for("user_management.register"))
        else:
            return render_template(
                "user_management/register.html",
                **Config.TEMPLATE_EXTRAS["register"]()
            )

    @blueprint.route("/confirm", methods=["GET"])
    def confirm() -> Union[Response, str]:
        """
        Confirms a user
        :return: The response
        """
        user_id = int(request.args["user_id"])
        confirm_key = request.args["confirm_key"]
        user: User = User.query.get(user_id)

        if user is None:
            flash(Config.STRINGS["user_does_not_exist"], "danger")
        elif user.confirmed:
            flash(Config.STRINGS["user_already_confirmed"], "warning")
        elif not user.verify_confirmation(confirm_key):
            flash(Config.STRINGS["confirmation_key_invalid"], "warning")
        else:
            user.confirmed = True
            db.session.commit()
            flash(Config.STRINGS["user_confirmed_successfully"], "success")
        return redirect(url_for("static.index"))

    @blueprint.route("/forgot", methods=["POST", "GET"])
    def forgot() -> Union[Response, str]:
        """
        Allows a user to reset their password
        :return: The response
        """
        if request.method == "POST":
            email = request.form["email"]
            recaptcha_result = verify_recaptcha(
                request.remote_addr,
                request.form.get("g-recaptcha-response", ""),
                Config.RECAPTCHA_SECRET_KEY
            )
            user: User = User.query.filter_by(email=email).first()

            if not recaptcha_result:
                flash(Config.STRINGS["recaptcha_incorrect"], "danger")
                return redirect(url_for("user_management.forgot"))
            else:
                if user is None:
                    # Fail silently to ensure that a potential attacker can't
                    # use the response to figure out information
                    # on registered users
                    pass
                else:
                    new_pass = generate_random(20)
                    user.password_hash = generate_hash(new_pass)
                    db.session.commit()

                    email_msg = render_template(
                        "email/forgot_password.html",
                        domain_name=Config.DOMAIN_NAME,
                        host_url=Config.base_url(),
                        target_url=os.path.join(Config.base_url(), "login"),
                        password=new_pass,
                        username=user.username,
                        **Config.TEMPLATE_EXTRAS["forgot_email"]()
                    )
                    send_email(
                        email,
                        Config.STRINGS["password_reset_email_title"],
                        email_msg,
                        Config.SMTP_HOST,
                        Config.SMTP_ADDRESS,
                        Config.SMTP_PASSWORD,
                        Config.SMTP_PORT
                    )
                flash(Config.STRINGS["password_was_reset"], "success")
                return redirect(url_for("static.index"))

        else:
            return render_template(
                "user_management/forgot.html",
                **Config.TEMPLATE_EXTRAS["forgot"]()
            )

    @blueprint.route("/profile", methods=["GET"])
    @login_required
    def profile() -> Union[Response, str]:
        """
        Allows a user to edit their profile details
        :return: The response
        """
        chat_id = TelegramChatId.query.filter_by(user=current_user).first()
        return render_template(
            "user_management/profile.html",
            **Config.TEMPLATE_EXTRAS["profile"](),
            telegram_chat_id=chat_id
        )

    @blueprint.route("/change_password", methods=["POST"])
    @login_required
    def change_password() -> Union[Response, str]:
        """
        Allows the user to change their password
        :return: The response
        """
        old_password = request.form["old_password"]
        new_password = request.form["new_password"]
        password_repeat = request.form["password_repeat"]
        user: User = current_user

        if new_password != password_repeat:
            flash(Config.STRINGS["passwords_do_not_match"], "danger")
        elif not user.verify_password(old_password):
            flash(Config.STRINGS["invalid_password"], "danger")
        else:
            user.password_hash = generate_hash(new_password)
            db.session.commit()
            flash(Config.STRINGS["password_changed"], "success")
        return redirect(url_for("user_management.profile"))

    @blueprint.route("/delete_user", methods=["POST"])
    @login_required
    def delete_user() -> Union[Response, str]:
        """
        Allows a user to delete their account
        :return: The response
        """
        password = request.form["password"]
        user: User = current_user

        if not user.verify_password(password):
            flash(Config.STRINGS["invalid_password"], "danger")
        else:
            app.logger.info("Deleting user {}".format(user))
            db.session.delete(user)
            db.session.commit()
            logout_user()
            flash(Config.STRINGS["user_was_deleted"], "success")
            return redirect(url_for("static.index"))
        return redirect(url_for("user_management.profile"))

    @blueprint.route("/register_telegram", methods=["POST"])
    @login_required
    def register_telegram() -> Union[Response, str]:
        """
        Allows the user to register a telegram chat ID
        :return: The response
        """
        telegram_chat_id = request.form["telegram_chat_id"]
        user: User = current_user
        chat_id = TelegramChatId.query.filter_by(user=user).first()

        if chat_id is None:
            chat_id = TelegramChatId(user=user, chat_id=telegram_chat_id)
            db.session.add(chat_id)
        else:
            chat_id.chat_id = telegram_chat_id
        db.session.commit()

        flash(Config.STRINGS["telegram_chat_id_set"], "success")
        chat_id.send_message(Config.STRINGS["telegram_chat_id_set"])
        return redirect(url_for("user_management.profile"))

    return blueprint



"""LICENSE
Copyright 2019 Hermann Krumrey <hermann@krumreyh.com>

This file is part of puffotter.

puffotter is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

puffotter is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with puffotter.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

from enum import Enum
from typing import Dict, Any, Type, Optional, List
from puffotter.flask.base import db
from sqlalchemy.inspection import inspect


class ModelMixin:
    """
    A mixin class that specifies a couple of methods all database
    models should implement
    """

    id = db.Column(
        db.Integer, primary_key=True, nullable=False, autoincrement=True
    )
    """
    The ID is the primary key of the table and increments automatically
    """

    def __json__(
            self,
            include_children: bool = False,
            ignore_keys: Optional[List[str]] = None
    ) -> Dict[str, Any]:
        """
        Generates a dictionary containing the information of this model
        :param include_children: Specifies if children data models will be
                                 included or if they're limited to IDs
        :param ignore_keys: If provided, will not include any of these keys
        :return: A dictionary representing the model's values
        """
        if ignore_keys is None:
            ignore_keys = []

        json_dict = {}

        relations: Dict[str, Type] = {
            key: value.mapper.class_
            for key, value in inspect(self.__class__).relationships.items()
        }

        for attribute in inspect(self).attrs:
            key = attribute.key
            value = attribute.value
            relation_cls = relations.get(key)

            if key in ignore_keys:
                continue
            elif key.endswith("_hash"):  # Skip password hashes etc
                continue
            elif isinstance(value, Enum):
                value = value.name
            elif relation_cls is not None and \
                    issubclass(relation_cls, ModelMixin):

                recursion_keys = []
                other_relations = \
                    list(inspect(relation_cls).relationships.values())
                for other_relation in other_relations:
                    other_relation_cls = other_relation.mapper.class_
                    if other_relation_cls == self.__class__:
                        recursion_keys.append(other_relation.key)
                recursion_keys += ignore_keys

                if include_children and value is not None:
                    if isinstance(value, list):
                        value = [
                            x.__json__(include_children, recursion_keys)
                            for x in value
                        ]
                    else:
                        value = value.__json__(
                            include_children, recursion_keys
                        )
                elif include_children and value is None:
                    value = None
                else:
                    continue

            json_dict[attribute.key] = value

        return json_dict

    def __str__(self) -> str:
        """
        :return: The string representation of this object
        """
        data = self.__json__()
        _id = data.pop("id")
        return "{}:{} <{}>".format(self.__class__.__name__, _id, str(data))

    def __repr__(self) -> str:
        """
        :return: A string with which the object may be generated
        """
        params = ""
        json_repr = self.__json__()

        enums = {}
        for key in json_repr:
            attr = getattr(self, key)
            if isinstance(attr, Enum):
                enum_cls = attr.__class__.__name__
                enum_val = attr.name
                enums[key] = "{}.{}".format(enum_cls, enum_val)

        for key, val in self.__json__().items():
            repr_arg = enums.get(key, repr(val))
            params += "{}={}, ".format(key, repr_arg)

        params = params.rsplit(",", 1)[0]

        return "{}({})".format(self.__class__.__name__, params)

    def __eq__(self, other: Any) -> bool:
        """
        Checks the model object for equality with another object
        :param other: The other object
        :return: True if the objects are equal, False otherwise
        """
        if "__json__" in dir(other):
            return other.__json__() == self.__json__()
        else:
            return False  # pragma: no cover

    def __hash__(self) -> int:
        """
        Creates a hash so that the model objects can be used as keys
        :return: None
        """
        return hash(repr(self))



"""LICENSE
Copyright 2019 Hermann Krumrey <hermann@krumreyh.com>

This file is part of puffotter.

puffotter is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

puffotter is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with puffotter.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

import logging
from colorama import Back, Fore, Style
from typing import Optional


class ColorLogger:

    def __init__(
            self,
            logger: logging.Logger,
            debug_bg: str = Back.WHITE,
            debug_fg: str = Fore.BLACK,
            info_bg: str = Back.BLACK,
            info_fg: str = Fore.WHITE,
            warning_bg: str = Back.YELLOW,
            warning_fg: str = Fore.BLACK,
            error_bg: str = Back.LIGHTRED_EX,
            error_fg: str = Fore.BLACK
    ):
        self.logger = logger
        self.debug_style = debug_bg + debug_fg
        self.info_style = info_bg + info_fg
        self.warning_style = warning_bg + warning_fg
        self.error_style = error_bg + error_fg

    def __format_message(
            self,
            level: int,
            message: str,
            bg: Optional[str],
            fg: Optional[str]
    ):
        """
        Formats a message
        :param level: The message's level
        :param message: The message to format
        :param bg: Optional background style override
        :param fg: Optional foreground style override
        :return: The formatted message
        """
        style = ""
        if bg is not None:
            style += bg
        if fg is not None:
            style += fg
        if style == "":
            style = {
                logging.DEBUG: self.debug_style,
                logging.INFO: self.info_style,
                logging.WARNING: self.warning_style,
                logging.ERROR: self.error_style
            }[level]
        return style + message + Style.RESET_ALL

    def debug(
            self,
            message: str,
            bg: Optional[str] = None,
            fg: Optional[str] = None
    ):
        """
        Logs a message at DEBUG level
        :param message: The message to log
        :param bg: Overrides the default background style
        :param fg: Overrides the default foreground style
        :return: None
        """
        self.logger.debug(
            self.__format_message(logging.DEBUG, message, bg, fg)
        )

    def info(
            self,
            message: str,
            bg: Optional[str] = None,
            fg: Optional[str] = None
    ):
        """
        Logs a message at INFO level
        :param message: The message to log
        :param bg: Overrides the default background style
        :param fg: Overrides the default foreground style
        :return: None
        """
        self.logger.info(self.__format_message(logging.INFO, message, bg, fg))

    def warning(
            self,
            message: str,
            bg: Optional[str] = None,
            fg: Optional[str] = None
    ):
        """
        Logs a message at WARNING level
        :param message: The message to log
        :param bg: Overrides the default background style
        :param fg: Overrides the default foreground style
        :return: None
        """
        self.logger.warning(
            self.__format_message(logging.WARNING, message, bg, fg)
        )

    def error(
            self,
            message: str,
            bg: Optional[str] = None,
            fg: Optional[str] = None
    ):
        """
        Logs a message at ERROR level
        :param message: The message to log
        :param bg: Overrides the default background style
        :param fg: Overrides the default foreground style
        :return: None
        """
        self.logger.error(
            self.__format_message(logging.ERROR, message, bg, fg)
        )



"""LICENSE
Copyright 2017 Hermann Krumrey <hermann@krumreyh.com>

This file is part of bundesliga-tippspiel.

bundesliga-tippspiel is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

bundesliga-tippspiel is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with bundesliga-tippspiel.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

# noinspection PyProtectedMember
from unittest import TestCase
from puffotter.crypto import generate_hash, generate_random, verify_password


class TestCrypto(TestCase):
    """
    Tests cryptographical functions
    """

    def test_hashing(self):
        """
        Tests that passwords can be hashed successfully
        :return: None
        """
        password_one = generate_random(30)
        password_two = generate_random(20)

        hash_one = generate_hash(password_one)
        hash_two = generate_hash(password_two)

        self.assertEqual(len(hash_one), len(hash_two))
        self.assertNotEqual(hash_one, hash_two)

        self.assertTrue(verify_password(password_one, hash_one))
        self.assertTrue(verify_password(password_two, hash_two))
        self.assertFalse(verify_password(password_one, hash_two))
        self.assertFalse(verify_password(password_two, hash_one))

    def test_salt(self):
        """
        Tests that password hashes are salted
        :return: None
        """
        password = generate_random(30)
        hash_one = generate_hash(password)
        hash_two = generate_hash(password)

        self.assertNotEqual(hash_one, hash_two)

        self.assertTrue(verify_password(password, hash_one))
        self.assertTrue(verify_password(password, hash_two))

    def test_hashing_strings(self):
        """
        Tests that password hashing and verifying works with string as well
        :return: None
        """
        password = str(generate_random(30))
        _hash = generate_hash(password)
        self.assertTrue(verify_password(password, _hash))

    def test_verifying_with_invalid_hash(self):
        """
        Tests that attempting to verify a password with an incorrectly
        formatted hash will return False
        :return: None
        """
        self.assertFalse(verify_password("A", "A"))



"""LICENSE
Copyright 2019 Hermann Krumrey <hermann@krumreyh.com>

This file is part of puffotter.

puffotter is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

puffotter is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with puffotter.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

import sys
from typing import IO, Optional
from colorama import Style, Fore, Back


def pprint(
        *objects: str,
        sep: str = " ",
        end: str = "\n",
        file: IO = sys.stdout,
        flush: bool = False,
        fg: Optional[str] = None,
        bg: Optional[str] = None
):
    """
    Function that extends the print function
    :return: None
    """
    code = ""
    if fg is not None:
        code += __convert_color_to_code(fg, "fore")
    if bg is not None:
        code += __convert_color_to_code(bg, "back")

    if code != "":
        object_list = []
        for i in range(0, len(objects)):
            object_list.append(code + str(objects[i]) + Style.RESET_ALL)
    else:
        object_list = list(objects)

    print(*object_list, sep=sep, end=end, file=file, flush=flush)


def __convert_color_to_code(color_string: str, mode: str) -> str:
    """
    Converts a color string into the correct ANSI code
    If the provided string does not match one of the predefined color strings,
    the strng will be returned as is.
    :param color_string: The string to convert
    :param mode: The mode, either 'fore' or 'back'.
                 Otherwise a ValueError will be raised.
    :return: The ANSI code
    """
    color_chart = {
        "lred": "LIGHTRED_EX",
        "red": "RED",
        "lyellow": "LIGHTYELLOW_EX",
        "yellow": "YELLOW",
        "lblue": "LIGHTBLUE_EX",
        "blue": "BLUE",
        "lgreen": "LIGHTGREEN_EX",
        "green": "GREEN",
        "lmagenta": "LIGHTMAGENTA_EX",
        "magenta": "MAGENTA",
        "lcyan": "LIGHTCYAN_EX",
        "cyan": "CYAN",
        "lwhite": "LIGHTWHITE_EX",
        "white": "WHITE",
        "lblack": "LIGHTBLACK_EX",
        "black": "BLACK"
    }
    code = color_chart.get(color_string.lower())
    if code is None:
        return color_string
    elif mode == "fore":
        return getattr(Fore, code)
    elif mode == "back":
        return getattr(Back, code)
    else:
        raise ValueError("Invalid Mode")



"""LICENSE
Copyright 2019 Hermann Krumrey <hermann@krumreyh.com>

This file is part of puffotter.

puffotter is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

puffotter is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with puffotter.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

import sentry_sdk
from functools import wraps
from typing import Callable
from flask import jsonify, make_response, request
from werkzeug.exceptions import Unauthorized
from puffotter.flask.exceptions import ApiException
from puffotter.flask.base import app


def api(func: Callable) -> Callable:
    """
    Decorator that handles common API patterns and ensures that
    the JSON response will always follow a certain pattern
    :param func: The function to wrap
    :return: The wrapper function
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        """
        Tries running the function and checks for errors
        :param args: args
        :param kwargs: kwargs
        :return: The JSON response including an appropriate HTTP status code
        """
        code = 200
        response = {"status": "ok"}

        try:
            is_json = request.content_type is not None \
                and request.content_type.startswith("application/json") \
                and request.is_json \
                and isinstance(request.get_json(silent=True), dict)
            if request.method in ["POST", "PUT", "DELETE"] and not is_json:
                raise ApiException("not in json format", 400)

            try:
                response["data"] = func(*args, **kwargs)
            except BaseException as e:
                for error_type in [
                    KeyError, TypeError, ValueError, ApiException
                ]:
                    app.logger.error("Caught exception in API: {}".format(e))
                    sentry_sdk.capture_exception(e)
                    if isinstance(e, error_type):
                        raise e
                raise ApiException("server error", 500)

        except (KeyError, TypeError, ValueError, ApiException) as e:

            response["status"] = "error"

            if isinstance(e, ApiException):
                code = e.status_code
                response["reason"] = e.reason

            else:
                code = 400
                response["reason"] = "bad request: {}".format(type(e).__name__)

        return make_response(jsonify(response), code)
    return wrapper


def api_login_required(func: Callable) -> Callable:
    """
    Decorator to make unauthorized API calls respond with JSON properly
    :param func: The function to wrap
    :return: The wrapped function
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        """
        Checks if flask-login throws an Unauthorized exception. If so,
        re-wrap the response in JSON
        :param args: The function arguments
        :param kwargs: The function keyword arguments
        :return: The newly wrapped response,
                 or just the plain response if authorized
        """

        try:
            resp = func(*args, **kwargs)
            return resp
        except Unauthorized:
            return make_response(
                jsonify({"status": "error", "reason": "unauthorized"}), 401
            )

    return wrapper



"""LICENSE
Copyright 2019 Hermann Krumrey <hermann@krumreyh.com>

This file is part of puffotter.

puffotter is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

puffotter is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with puffotter.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""


from flask import Flask
from flask_sqlalchemy import SQLAlchemy
from flask_login import LoginManager

app = Flask(__name__)
"""
The Flask App
"""

db = SQLAlchemy()
"""
The SQLAlchemy database connection
"""

login_manager = LoginManager(app)
"""
The Flask-Login Login Manager
"""



"""LICENSE
Copyright 2020 Hermann Krumrey <hermann@krumreyh.com>

This file is part of fat-ffipd.

fat-ffipd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

fat-ffipd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with fat-ffipd.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

from unittest.mock import patch
from puffotter.flask.test.TestFramework import _TestFramework


class TestForgotRoute(_TestFramework):
    """
    Class that tests password reset features
    """

    def test_page_get(self):
        """
        Tests getting the page
        :return: None
        """
        resp = self.client.get("/forgot")
        self.assertTrue(b"<!--user_management/forgot.html-->" in resp.data)

    def test_resetting_password(self):
        """
        Tests successfully resetting a password
        :return: None
        """
        user, password, _ = self.generate_sample_user()
        with self.client:
            with patch("puffotter.flask.routes.user_management.send_email") \
                    as m:
                with patch("puffotter.flask.routes.user_management."
                           "generate_random", lambda x: "testpass"):
                    self.assertEqual(0, m.call_count)
                    resp = self.client.post(
                        "/forgot",
                        follow_redirects=True,
                        data={
                            "email": user.email,
                            "g-recaptcha-response": ""
                        }
                    )
                    self.assertEqual(1, m.call_count)

            expected = self.config.STRINGS["password_was_reset"]
            self.assertTrue(expected.encode("utf-8") in resp.data)
            self.assertTrue(b"<!--static/index.html-->" in resp.data)
            self.assertFalse(user.verify_password(password))
            self.assertTrue(user.verify_password("testpass"))

    def test_unsuccessfully_resetting_password(self):
        """
        Tests unsuccessfully resetting a password
        :return: None
        """
        user, password, _ = self.generate_sample_user()
        with self.client:
            with patch("puffotter.flask.routes.user_management.send_email") \
                    as m:
                with patch("puffotter.flask.routes.user_management."
                           "generate_random", lambda x: "testpass"):
                    self.assertEqual(0, m.call_count)
                    resp = self.client.post(
                        "/forgot",
                        follow_redirects=True,
                        data={
                            "email": user.email + "AAA",
                            "g-recaptcha-response": ""
                        }
                    )
                    self.assertEqual(0, m.call_count)

            expected = self.config.STRINGS["password_was_reset"]
            self.assertTrue(expected.encode("utf-8") in resp.data)
            self.assertTrue(b"<!--static/index.html-->" in resp.data)
            self.assertTrue(user.verify_password(password))
            self.assertFalse(user.verify_password("testpass"))

    def test_invalid_recaptcha(self):
        """
        Tests that invalid ReCaptcha responses are handled correctly
        :return: None
        """
        user, password, _ = self.generate_sample_user()
        with self.client:
            with patch("puffotter.flask.routes.user_management.send_email") \
                    as m:
                with patch("puffotter.flask.routes.user_management"
                           ".generate_random", lambda x: "testpass"):
                    with patch("puffotter.flask.routes.user_management"
                               ".verify_recaptcha",
                               lambda x, y, z: False):
                        self.assertEqual(0, m.call_count)
                        resp = self.client.post(
                            "/forgot",
                            follow_redirects=True,
                            data={
                                "email": user.email,
                                "g-recaptcha-response": ""
                            }
                        )
                        self.assertEqual(0, m.call_count)

            expected = self.config.STRINGS["recaptcha_incorrect"]\
                .encode("utf-8")
            self.assertTrue(expected in resp.data)
            self.assertTrue(b"<!--user_management/forgot.html-->" in resp.data)
            self.assertTrue(user.verify_password(password))
            self.assertFalse(user.verify_password("testpass"))
